//| Color chooser for the Fast Light Tool Kit (FLTK).

#include <FL/Fl.H>
#include <FL/Fl_RGBA_Chooser.H>
#include <FL/fl_draw.H>
#include <FL/math.h>
#include <stdio.h>

// Besides being a useful object on it's own, the Fl_RGBA_Chooser was
// an attempt to make a complex composite object that could be easily
// imbedded into a user interface.  If you wish to make complex objects
// of your own, be sure to read this code.

// The function Fl_RGBA_Chooser() creates a window containing a color
// chooser and a few buttons and current-color indicators.  It is an
// easier interface for simple programs that just need a color.

// The "hue box" can be a circle or rectilinear.
// You get a circle by defining this:
//#define CIRCLE 1
// And the "hue box" can auto-update when the value changes
// you get this by defining this:
#define UPDATE_HUE_BOX 1

void Fl_RGBA_Chooser::hsv2rgb(
	float H, float S, float V, float& r, float& g, float& b) {
  if (S < 5.0e-6) {
    r = g = b = V;
  } else {
    int i = (int)H;  
    float f = H - (float)i;
    float p1 = V*(1.0-S);
    float p2 = V*(1.0-S*f);
    float p3 = V*(1.0-S*(1.0-f));
    switch (i) {
    case 0: r = V;   g = p3;  b = p1;  break;
    case 1: r = p2;  g = V;   b = p1;  break;
    case 2: r = p1;  g = V;   b = p3;  break;
    case 3: r = p1;  g = p2;  b = V;   break;
    case 4: r = p3;  g = p1;  b = V;   break;
    case 5: r = V;   g = p1;  b = p2;  break;
    }
  }
}

void Fl_RGBA_Chooser::rgb2hsv(
	float r, float g, float b, float& H, float& S, float& V) {
  float maxv = r > g ? r : g; if (b > maxv) maxv = b;
  V = maxv;
  if (maxv>0) {
    float minv = r < g ? r : g; if (b < minv) minv = b;
    S = 1.0 - float(minv)/maxv;
    if (maxv > minv) {
      if (maxv == r) {H = (g-b)/float(maxv-minv); if (H<0) H += 6.0;}
      else if (maxv == g) H = 2.0+(b-r)/float(maxv-minv);
      else H = 4.0+(r-g)/float(maxv-minv);
    }
  }
}

enum {M_RGB, M_BYTE, M_HEX, M_HSV}; // modes
static Fl_Menu_Item mode_menu[] = {
  {"rgb"},
  {"byte"},
  {"hex"},
  {"hsv"},
  {0},
  {"rgba"},
  {"byte"},
  {"hex"},
  {"hsva"},
  {0}
};

int Flcc_Value_Input_A::format(char* buf) {
  Fl_RGBA_Chooser* c = (Fl_RGBA_Chooser*)parent();
  if (c->mode() == M_HEX) return sprintf(buf,"0x%02X", int(value()));
  else return Fl_Valuator::format(buf);
}

void Fl_RGBA_Chooser::set_valuators() {
  switch (mode()) {
  case M_RGB:
    rvalue.range(0,1); rvalue.step(1,1000); rvalue.value(r_);
    gvalue.range(0,1); gvalue.step(1,1000); gvalue.value(g_);
    bvalue.range(0,1); bvalue.step(1,1000); bvalue.value(b_);
    avalue.range(0,1); avalue.step(1,1000); avalue.value(a_);
    break;
  case M_BYTE:
  case M_HEX:
    rvalue.range(0,255); rvalue.step(1); rvalue.value(int(255*r_+.5));
    gvalue.range(0,255); gvalue.step(1); gvalue.value(int(255*g_+.5));
    bvalue.range(0,255); bvalue.step(1); bvalue.value(int(255*b_+.5));
    avalue.range(0,255); avalue.step(1); avalue.value(int(255*a_+.5));
    break;
  case M_HSV:
    rvalue.range(0,6); rvalue.step(1,1000); rvalue.value(hue_);
    gvalue.range(0,1); gvalue.step(1,1000); gvalue.value(saturation_);
    bvalue.range(0,1); bvalue.step(1,1000); bvalue.value(value_);
    avalue.range(0,1); avalue.step(1,1000); avalue.value(a_);
    break;
  }
}

int Fl_RGBA_Chooser::rgb(float r, float g, float b, float a) {
  if (a == -1) a = a_;
  if (r == r_ && g == g_ && b == b_ && a == a_) return 0;
  float pa = a_;
  r_ = r; g_ = g; b_ = b; a_ = a;
  float ph = hue_;
  float ps = saturation_;
  float pv = value_;
  rgb2hsv(r,g,b,hue_,saturation_,value_);
  set_valuators();
  if (value_ != pv) {
#ifdef UPDATE_HUE_BOX
    huebox.damage(FL_DAMAGE_SCROLL);
#endif
    valuebox.damage(FL_DAMAGE_EXPOSE);}
  if (hue_ != ph || saturation_ != ps) {
    huebox.damage(FL_DAMAGE_EXPOSE); 
    valuebox.damage(FL_DAMAGE_SCROLL);
  }
  if (a_ != pa) alphabox.damage(FL_DAMAGE_EXPOSE);
  if (hue_ != ph || saturation_ != ps || value_ != pv)
      alphabox.damage(FL_DAMAGE_SCROLL);
  return 1;
}

int Fl_RGBA_Chooser::hsv(float h, float s, float v, float a) {
  if (a == -1) a = a_;
  float pa = a_;
  h = fmod(h,6.0); if (h < 0.0) h += 6.0;
  if (s < 0.0) s = 0.0; else if (s > 1.0) s = 1.0;
  if (v < 0.0) v = 0.0; else if (v > 1.0) v = 1.0;
  if (h == hue_ && s == saturation_ && v == value_ && a == a_) return 0;
  float ph = hue_;
  float ps = saturation_;
  float pv = value_;
  hue_ = h; saturation_ = s; value_ = v; a_ = a;
  if (value_ != pv) {
#ifdef UPDATE_HUE_BOX
    huebox.damage(FL_DAMAGE_SCROLL);
#endif
    valuebox.damage(FL_DAMAGE_EXPOSE);}
  if (hue_ != ph || saturation_ != ps) {
    huebox.damage(FL_DAMAGE_EXPOSE); 
    valuebox.damage(FL_DAMAGE_SCROLL);
  }
  if (a_ != pa) alphabox.damage(FL_DAMAGE_EXPOSE);
  if (hue_ != ph || saturation_ != ps || value_ != pv)
      alphabox.damage(FL_DAMAGE_SCROLL);
  hsv2rgb(h,s,v,r_,g_,b_);
  set_valuators();
  return 1;
}

////////////////////////////////////////////////////////////////

void Fl_RGBA_Chooser::rgb_cb(Fl_Widget* o, void*) {
  Fl_RGBA_Chooser* c = (Fl_RGBA_Chooser*)(o->parent());
  float r = c->rvalue.value();
  float g = c->gvalue.value();
  float b = c->bvalue.value();
  float a = c->avalue.value();
  if (c->mode() == M_HSV) {
    if (c->hsv(r,g,b,a)) c->set_ok_butt();
    return;
  }
  if (c->mode() != M_RGB) {
    r = r/255;
    g = g/255;
    b = b/255;
    a = a/255;
  }
  if (c->rgb(r,g,b,a)) c->set_ok_butt();
}

void Fl_RGBA_Chooser::mode_cb(Fl_Widget* o, void*) {
  Fl_RGBA_Chooser* c = (Fl_RGBA_Chooser*)(o->parent());
  // force them to redraw even if value is the same:
  c->rvalue.value(-1);
  c->gvalue.value(-1);
  c->bvalue.value(-1);
  c->avalue.value(-1);
  c->set_valuators();
}

/**************************************************************************/
/**************************************************************************/

void Fl_RGBA_Chooser::set_master() {
  master->set_rgba(&ok_color); master->do_callback(master, &ok_color);
}

void Fl_RGBA_Chooser::undo() {
  rgb(cc_color.r, cc_color.g, cc_color.b, cc_color.a);
  set_ok_butt();
}

void Fl_RGBA_Chooser::xit() {
  master->wipe_chooser();
}

/***************************************/

void Fl_RGBA_Chooser::ok_cb(Fl_Button* b, Fl_RGBA_Chooser* c) {
  c->set_master(); c->xit(); delete c;
} 

void Fl_RGBA_Chooser::set_cb(Fl_Button* b, Fl_RGBA_Chooser* c)  { c->set_master(); }


void Fl_RGBA_Chooser::undo_cb(Fl_Button* b, Fl_RGBA_Chooser* c) { c->undo(); }

void Fl_RGBA_Chooser::xit_cb(Fl_Button* b, Fl_RGBA_Chooser* c)  { c->xit(); delete c; }

void Fl_RGBA_Chooser::set_ok_butt() {
  ok_color.set_rgba(r_, g_, b_, a_);
  ok_color.damage(FL_DAMAGE_EXPOSE);
}

void Fl_RGBA_Chooser::TraceManager(Fl_Widget* w) {
  Fl_SWM_Client::TraceManager(w);
  if(swm_manager==0) { // resize to reasnable defaults
    Fl_Scaled_Windows_Manager::ResizeGroup(this, 1, 1, 8, 20, 14);
  }
}

////////////////////////////////////////////////////////////////

Fl_RGBA_Chooser::Fl_RGBA_Chooser(Fl_RGBA_Button* w, const char* l)
  : Fl_Window(28,8,l), Fl_SWM_Client(), master(w),
    huebox	( 0, 0,18, 7),
    valuebox	(18, 0, 2, 7),
    alphabox	(20, 0, 2, 7),
    choice	(22, 0, 6, 1),
    rvalue	(22, 1, 6, 1),
    gvalue	(22, 2, 6, 1),
    bvalue	(22, 3, 6, 1),
    avalue	(22, 4, 6, 1),
    ok_butt	( 0, 7, 7, 1, "OK"),
    ok_color	( 7, 7, 7, 1, "Set"),
    cc_color	(14, 7, 7, 1, "Undo"),
    cc_butt	(21, 7, 7, 1, "Xit"),
    resize_box	( 1, 5, 5, 2)
{
  swm_size_range = new SWM_Size_Range(28, 8, 28*4, 8*4);
  resizable(&resize_box);

  r_ = g_ = b_ = 0.0; a_ = 1.0;
  hue_ = 0.0;
  saturation_ = 0.0;
  value_ = 0.0;
  huebox.box(FL_DOWN_FRAME);
  valuebox.box(FL_DOWN_FRAME);
  alphabox.box(FL_DOWN_FRAME);
  choice.menu(mode_menu);
  set_valuators();
  rvalue.callback(rgb_cb);
  gvalue.callback(rgb_cb);
  bvalue.callback(rgb_cb);
  avalue.callback(rgb_cb);
  choice.callback(mode_cb);
  choice.box(FL_THIN_UP_BOX);
  choice.textfont(FL_HELVETICA_BOLD_ITALIC);

  ok_butt.callback((Fl_Callback*)ok_cb, this);

  ok_color.box(FL_EMBOSSED_FRAME); ok_color.set_rgba(w);
  ok_color.callback((Fl_Callback*)set_cb, this);

  cc_color.box(FL_EMBOSSED_FRAME); cc_color.set_rgba(w);
  cc_color.callback((Fl_Callback*)undo_cb, this);

  cc_butt.callback((Fl_Callback*)xit_cb, this);

  rgb(w->r, w->g, w->b, w->a);

  end();
}

/**************************************************************************/
/**************************************************************************/
// Flcc classes
/**************************************************************************/

static void tohs(float x, float y, float& h, float& s) {
#ifdef CIRCLE
  x = 2*x-1;
  y = 1-2*y;
  s = sqrt(x*x+y*y); if (s > 1.0) s = 1.0;
  h = (3.0/M_PI)*atan2(y,x);
  if (h<0) h += 6.0;
#else
  h = fmod(6.0*x,6.0); if (h < 0.0) h += 6.0;
  s = 1.0-y; if (s < 0.0) s = 0.0; else if (s > 1.0) s = 1.0;
#endif
}

int Flcc_HueBox_A::handle(int e) {
  static float ih, is;
  Fl_RGBA_Chooser* c = (Fl_RGBA_Chooser*)parent();
  switch (e) {
  case FL_PUSH:
    ih = c->hue();
    is = c->saturation();
  case FL_DRAG: {
    float Xf, Yf, H, S;
    Xf = (Fl::event_x()-x()-Fl::box_dx(box()))/float(w()-Fl::box_dw(box()));
    Yf = (Fl::event_y()-y()-Fl::box_dy(box()))/float(h()-Fl::box_dh(box()));
    tohs(Xf, Yf, H, S);
    if (fabs(H-ih) < 3*6.0/w()) H = ih;
    if (fabs(S-is) < 3*1.0/h()) S = is;
    if (Fl::event_state(FL_CTRL)) H = ih;
    if (c->hsv(H, S, c->value())) c->set_ok_butt();
    } return 1;
  default:
    return 0;
  }
}

static void generate_image(void* vv, int X, int Y, int W, uchar* buf) {
  Flcc_HueBox_A* v = (Flcc_HueBox_A*)vv;
  int iw = v->w()-Fl::box_dw(v->box());
  float Yf = float(Y)/(v->h()-Fl::box_dh(v->box()));
#ifdef UPDATE_HUE_BOX
  const float V = ((Fl_RGBA_Chooser*)(v->parent()))->value();
#else
  const float V = 1.0;
#endif
  for (int x = X; x < X+W; x++) {
    float Xf = float(x)/iw;
    float H,S; tohs(Xf,Yf,H,S);
    float r,g,b;
    Fl_RGBA_Chooser::hsv2rgb(H,S,V,r,g,b);
    *buf++ = uchar(255*r+.5);
    *buf++ = uchar(255*g+.5);
    *buf++ = uchar(255*b+.5);
  }
}

void Flcc_HueBox_A::draw() {
  if (damage()&FL_DAMAGE_ALL) draw_box();
  int x1 = x()+Fl::box_dx(box());
  int y1 = y()+Fl::box_dy(box());
  int w1 = w()-Fl::box_dw(box());
  int h1 = h()-Fl::box_dh(box());
  if (damage() == FL_DAMAGE_EXPOSE) fl_clip(x1+px,y1+py,6,6);
  fl_draw_image(generate_image, this, x1, y1, w1, h1);
  if (damage() == FL_DAMAGE_EXPOSE) fl_pop_clip();
  Fl_RGBA_Chooser* c = (Fl_RGBA_Chooser*)parent();
#ifdef CIRCLE
  int X = int(.5*(cos(c->hue()*(M_PI/3.0))*c->saturation()+1) * (w1-6));
  int Y = int(.5*(1-sin(c->hue()*(M_PI/3.0))*c->saturation()) * (h1-6));
#else
  int X = int(c->hue()/6.0*(w1-6));
  int Y = int((1-c->saturation())*(h1-6));
#endif
  if (X < 0) X = 0; else if (X > w1-6) X = w1-6;
  if (Y < 0) Y = 0; else if (Y > h1-6) Y = h1-6;
  //  fl_color(c->value()>.75 ? FL_BLACK : FL_WHITE);
  draw_box(FL_UP_BOX,x1+X,y1+Y,6,6,FL_GRAY);
  px = X; py = Y;
}


////////////////////////////////////////////////////////////////

int Flcc_ValueBox_A::handle(int e) {
  static float iv;
  Fl_RGBA_Chooser* c = (Fl_RGBA_Chooser*)parent();
  switch (e) {
  case FL_PUSH:
    iv = c->value();
  case FL_DRAG: {
    float Yf;
    Yf = 1-(Fl::event_y()-y()-Fl::box_dy(box()))/float(h()-Fl::box_dh(box()));
    if (fabs(Yf-iv)<(3*1.0/h())) Yf = iv;
    if (c->hsv(c->hue(),c->saturation(),Yf)) c->set_ok_butt();
    } return 1;
  default:
    return 0;
  }
}

static float tr, tg, tb, ta;
static void generate_vimage(void* vv, int X, int Y, int W, uchar* buf) {
  Flcc_ValueBox_A* v = (Flcc_ValueBox_A*)vv;
  float Yf = 255*(1.0-float(Y)/(v->h()-Fl::box_dh(v->box())));
  uchar r = uchar(tr*Yf+.5);
  uchar g = uchar(tg*Yf+.5);
  uchar b = uchar(tb*Yf+.5);
  for (int x = X; x < X+W; x++) {
    *buf++ = r; *buf++ = g; *buf++ = b;
  }
}

void Flcc_ValueBox_A::draw() {
  if (damage()&FL_DAMAGE_ALL) draw_box();
  Fl_RGBA_Chooser* c = (Fl_RGBA_Chooser*)parent();
  c->hsv2rgb(c->hue(),c->saturation(),1.0,tr,tg,tb);
  int x1 = x()+Fl::box_dx(box());
  int y1 = y()+Fl::box_dy(box());
  int w1 = w()-Fl::box_dw(box());
  int h1 = h()-Fl::box_dh(box());
  if (damage() == FL_DAMAGE_EXPOSE) fl_clip(x1,y1+py,w1,6);
  fl_draw_image(generate_vimage, this, x1, y1, w1, h1);
  if (damage() == FL_DAMAGE_EXPOSE) fl_pop_clip();
  int Y = int((1-c->value()) * (h1-6));
  if (Y < 0) Y = 0; else if (Y > h1-6) Y = h1-6;
  draw_box(FL_UP_BOX,x1,y1+Y,w1,6,FL_GRAY);
  py = Y;
}

////////////////////////////////////////////////////////////////

int Flcc_AlphaBox_A::handle(int e) {
  static float iv;
  Fl_RGBA_Chooser* c = (Fl_RGBA_Chooser*)parent();
  switch (e) {
  case FL_PUSH:
    iv = c->a();
  case FL_DRAG: {
    float Yf;
    Yf = 1-(Fl::event_y()-y()-Fl::box_dy(box()))/float(h()-Fl::box_dh(box()));
    if (Yf<0.0) Yf = 0.0; if (Yf>1.0) Yf = 1.0; 
    if (c->rgb(c->r(),c->g(),c->b(),Yf)) c->set_ok_butt();
    } return 1;
  default:
    return 0;
  }
}

static void generate_aimage(void* vv, int X, int Y, int W, uchar* buf) {
  Flcc_ValueBox_A* v = (Flcc_ValueBox_A*)vv;
  float Yf = float(Y)/(v->h()-Fl::box_dh(v->box())), iYf = 1.0-Yf;
  uchar rg = (uchar)(255*(iYf*tr + Yf*0.8));
  uchar gg = (uchar)(255*(iYf*tg + Yf*0.4));
  uchar bg = (uchar)(255*(iYf*tb + Yf*0.4));
  uchar r, g, b;
  if ((Y&0x07)==4) {
    r = rg; g = gg; b = bg;
  } else {
    r = (uchar)(255*(iYf*tr + Yf*0.8));
    g = (uchar)(255*(iYf*tg + Yf*0.8));
    b = (uchar)(255*(iYf*tb + Yf*0.8));
  }
  for (int x = X; x < X+W; x++) {
    if ((x&0x07)==4) {
      *buf++ = rg; *buf++ = gg; *buf++ = bg;
    } else {
      *buf++ = r; *buf++ = g; *buf++ = b;
    }
  }
}

void Flcc_AlphaBox_A::draw() {
  if (damage()&FL_DAMAGE_ALL) draw_box();
  Fl_RGBA_Chooser* c = (Fl_RGBA_Chooser*)parent();
  tr=c->r(); tg=c->g();  tb=c->b();
  int x1 = x()+Fl::box_dx(box());
  int y1 = y()+Fl::box_dy(box());
  int w1 = w()-Fl::box_dw(box());
  int h1 = h()-Fl::box_dh(box());
  if (damage() == FL_DAMAGE_EXPOSE) fl_clip(x1,y1+py,w1,6);
  fl_draw_image(generate_aimage, this, x1, y1, w1, h1);
  if (damage() == FL_DAMAGE_EXPOSE) fl_pop_clip();
  int Y = int((1-c->a()) * (h1-6));
  if (Y < 0) Y = 0; else if (Y > h1-6) Y = h1-6;
  draw_box(FL_UP_BOX,x1,y1+Y,w1,6,FL_GRAY);
  py = Y;
}
